@@ -5,10 +5,11 @@ import logging
 import pathlib
 import re
 import time
-import typing
 from dataclasses import dataclass
 from decimal import Decimal
-from typing import Literal
+from typing import Literal, Any
+
+from .constants import RUST_I64_MIN, RUST_U64_MAX, RUST_I64_MAX, QUOTE_DECIMALS
 
 import base58
 from anchorpy import Provider, Wallet
@@ -41,9 +42,9 @@ from mango_explorer_v4.types import serum3_side, serum3_self_trade_behavior, ser
 from mango_explorer_v4.types import place_order_type
 
 
-# logging.basicConfig(
-#     level=logging.DEBUG
-# )
+logging.basicConfig(
+    level=logging.INFO
+)
 
 
 @dataclass
@@ -52,13 +53,14 @@ class MangoClient():
     mango_account_pk: str
     group_config: dict
     mango_account: MangoAccount
+    serum_market_configs: [dict]
+    perp_market_configs: [dict]
     serum_markets: [Serum3Market]
     serum_markets_external: [AsyncMarket]
     banks: [Bank]
     mint_infos: [MintInfo]
     perp_markets: [PerpMarket]
     rpc_url: str
-    symbols: [str]
 
     @staticmethod
     async def connect(
@@ -72,6 +74,8 @@ class MangoClient():
         rpc_url: str = 'https://mango.rpcpool.com/0f9acc0d45173b51bf7d7e09c1e5'
         # ^ Can use the default RPC endpoint or whichever so desired
     ):
+        # TODO: There's a bunch of asynchronous calls here, which could reasonably be parallelized
+        # to shorten the lib's init time, which is now sitting at around 2 seconds - annoying
         provider = Provider(
             AsyncClient(rpc_url, Processed),
             Wallet(
@@ -92,16 +96,23 @@ class MangoClient():
         if mango_account.owner != provider.wallet.public_key:
             raise ValueError('Mango account is not owned by the secret key entered')
 
-        ids = json.loads(open(pathlib.Path(__file__).parent.parent / 'ids.json').read())
+        ids = json.loads(open(pathlib.Path(__file__).parent / 'ids.json').read())
+        # TODO: ^ Make this fetch from https://mango-transaction-log.herokuapp.com/v4/group-metadata instead
 
         group_config = [group_config for group_config in ids['groups'] if PublicKey(group_config['publicKey']) == mango_account.group][0]
 
+        perp_market_configs = [
+            perp_market_config for perp_market_config in group_config['perpMarkets'] if perp_market_config['active']
+        ]
+
+        serum_market_configs = [
+            serum_market_config for serum_market_config in group_config['serum3Markets'] if serum_market_config['active']
+        ]
+
         serum_markets = await Serum3Market.fetch_multiple(
             provider.connection,
             [
-                PublicKey(serum3_market_config['publicKey'])
-                for serum3_market_config in group_config['serum3Markets']
-                if serum3_market_config['active']
+                PublicKey(serum3_market_config['publicKey']) for serum3_market_config in serum_market_configs
             ]
         )
 
@@ -143,43 +154,61 @@ class MangoClient():
 
         perp_markets = await PerpMarket.fetch_multiple(
             provider.connection,
-            [perp_market['publicKey'] for perp_market in group_config['perpMarkets']]
+            [perp_market_config['publicKey'] for perp_market_config in perp_market_configs]
         )
 
-        symbols = [
-            *[perp_market_config['name'] for perp_market_config in group_config['perpMarkets']],
-            *[serum3_market_config['name'] for serum3_market_config in group_config['serum3Markets']]
-        ]
-
         return MangoClient(
             provider,
             mango_account_pk,
             mango_account=mango_account,
             group_config=group_config,
+            serum_market_configs=serum_market_configs,
+            perp_market_configs=perp_market_configs,
             serum_markets=serum_markets,
             serum_markets_external=serum_markets_external,
             banks=banks,
             mint_infos=mint_infos,
             perp_markets=perp_markets,
-            rpc_url=rpc_url,
-            symbols=symbols
+            rpc_url=rpc_url
         )
 
-    async def symbols(self):
-        # Can't make it static yet because need to know which group
-        # to use, and group is inferred from the Mango account
+    def symbols(self):
+        # Can't make it a static / class function yet because need to know
+        # which group to use, and group is inferred from the Mango account
+
+        # This might not be the best format for keeping symbols organized,
+        # as it presumes that perpetual and spot market names would never
+        # conflict, but it works for now
+
+        # TODO: Add minimum order size, tick size, lot size and liquidation fee
         return [
-            {
-                'name': market_config['name'],
-                'baseCurrency': market_config['name'].split('/')[0],
-                'quoteCurrency': market_config['name'].split('/')[1],
-                'makerFees': - (0.5 / 1e4),
-                'takerFees': (1 / 1e4)
-            }
-            for market_config in self.group_config['serum3Markets']
+            *[
+                {
+                    'name': perp_market_config['name'],
+                    'type': 'perpetual',
+                    'base_currency': perp_market_config['name'].split('-')[0],
+                    'quote_currency': perp_market_config['name'].split('-')[1],
+                    'maker_fees': - (1 / 1e4),
+                    'taker_fees': (4 / 1e4)
+                }
+                for perp_market_config in self.perp_market_configs
+            ],
+            *[
+                {
+                    'name': serum3_market_config['name'],
+                    'type': 'spot',
+                    'base_currency': serum3_market_config['name'].split('/')[0],
+                    'quote_currency': serum3_market_config['name'].split('/')[1],
+                    'maker_fees': - (0.5 / 1e4),
+                    'taker_fees': (1 / 1e4)
+                }
+                for serum3_market_config in self.serum_market_configs
+            ]
         ]
 
     async def orderbook_l2(self, symbol: str, depth: int = 100):
+        # TODO: Validate that the symbol entered is valid
+
         market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]
 
         match market_type:
@@ -220,15 +249,74 @@ class MangoClient():
 
                 return orderbook
             case 'perpetual':
+                @dataclass
+                class PerpOrder:
+                    seq_num: int
+                    order_id: int
+                    owner: PublicKey
+                    open_orders_slot: int
+                    fee_tier: 0
+                    ui_price: float
+                    price: int
+                    ui_size: float
+                    size: int
+                    side: Literal['bids', 'asks']
+                    timestamp: int
+                    expiry_timestamp: int
+                    perp_market_index: int
+                    is_expired: bool
+                    is_oracle_pegged: bool
+                    oracle_pegged_properties: Any
+
+
+                    @staticmethod
+                    def build(
+                        perp_market: PerpMarket,
+                        leaf_node: LeafNode,
+                        side: Literal['bids', 'asks'],
+                        is_oracle_pegged: bool = False
+                    ):
+                        if is_oracle_pegged:
+                            price_lots = leaf_node.key >> 64
+                        else:
+                            price_lots = leaf_node.key >> 64
+
+                        now = int(time.time())
+
+                        expiry_timestamp = leaf_node.timestamp + leaf_node.time_in_force if leaf_node.time_in_force else sys.maxsize
+
+                        is_expired = now > expiry_timestamp
+
+                        return PerpOrder(
+                            {
+                                'bids': RUST_U64_MAX - (leaf_node.key & ((1 << 64) - 1)),
+                                'asks': leaf_node.key & ((1 << 64) - 1)
+                            }[side],
+                            leaf_node.key,
+                            leaf_node.owner,
+                            leaf_node.owner_slot,
+                            0,
+                            float(Decimal(price_lots) * Decimal(perp_market.price_lots_to_ui_converter)),
+                            price_lots,
+                            float(Decimal(leaf_node.quantity) * Decimal(perp_market.base_lots_to_ui_converter)),
+                            leaf_node.quantity,
+                            side,
+                            leaf_node.timestamp,
+                            expiry_timestamp,
+                            perp_market.perp_market_index,
+                            is_expired,
+                            is_oracle_pegged,
+                            None
+                        )
+
+
                 perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]
 
                 perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]
 
                 [bids, asks] = await BookSide.fetch_multiple(self.provider.connection, [perp_market.bids, perp_market.asks])
 
-                def to_inner_node(data):
-                    pass
-
+                # TODO: Abstract away the pretty much the same body in fixed_items() and oracle_pegged_items()
                 def items(book_side: BookSide, side: Literal['bids', 'asks']):
                     def fixed_items():
                         if book_side.roots[0].leaf_count == 0:
@@ -236,31 +324,61 @@ class MangoClient():
 
                         stack = [book_side.roots[0].maybe_node]
 
+                        [left, right] = [1, 0] if side == 'bids' else [0, 1]
+
                         while len(stack) > 0:
                             index = stack.pop()
 
                             node = book_side.nodes.nodes[index]
 
-                            [left, right] = [1, 0] if side == 'bids' else [0, 1]
-
                             match node.tag:
                                 case 1:
-                                    inner_node = InnerNode.from_decoded([1] + node.data)
+                                    inner_node = InnerNode.layout.parse(bytes([1] + node.data))
 
                                     stack.extend([inner_node.children[right], inner_node.children[left]])
                                 case 2:
-                                    leaf_node = LeafNode.from_decoded([2] + node.data)
+                                    leaf_node: LeafNode = LeafNode.layout.parse(bytes([2] + node.data))
 
-                                    yield
+                                    yield PerpOrder.build(
+                                        perp_market,
+                                        leaf_node,
+                                        side
+                                    )
 
-                            yield node.tag
+                    def oracle_pegged_items():
+                        if book_side.roots[1].leaf_count == 0:
+                            return
 
-                    return fixed_items()
+                        stack = [book_side.roots[1].maybe_node]
 
+                        [left, right] = [1, 0] if side == 'bids' else [0, 1]
 
-                return list(items(asks, 'asks'))
+                        while len(stack) > 0:
+                            index = stack.pop()
 
+                            node = book_side.nodes.nodes[index]
 
+                            match node.tag:
+                                case 1:
+                                    inner_node = InnerNode.layout.parse(bytes([1] + node.data))
+
+                                    stack.extend([inner_node.children[right], inner_node.children[left]])
+                                case 2:
+                                    leaf_node: LeafNode = LeafNode.layout.parse(bytes([2] + node.data))
+
+                                    yield PerpOrder.build(
+                                        perp_market,
+                                        leaf_node,
+                                        side,
+                                        True
+                                    )
+
+                    return [
+                        list(fixed_items()),
+                        list(oracle_pegged_items())
+                    ]
+
+                return items(bids, 'bids')
 
     async def snapshots_l2(self, symbol: str, depth: int = 50):
         serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]
@@ -758,6 +876,7 @@ class MangoClient():
                 return response
 
     async def balances(self):
+        # TODO: Clean up this mess
         return [
             {
                 'symbol': meta['symbol'],
